

local BASE, VERSION, MINOR = "LibBuilder", "1.0", 1
local MAJOR = BASE.."-"..VERSION


local LibBuilder = LibStub:NewLibrary(MAJOR, MINOR)
if not LibBuilder then return end


LibBuilder.errorStrings = {
	SHELL = "%s: %s!", 
	BADVERSION = "NewLib requires a textual major and numeric minor version", 
	COLLISION = "Attempted to build %s, rev. %s before closing %s, rev. %s", 
	NOLIB = "%s called but no library is being built", 
	LEFTOPEN = "Library %s, rev. %s was not closed properly",
}


--	Create a frame so we can throw a warning if 
--	a library is still open when OnUpdate fires.
--	Use OnUpdate instead of ADDON_LOADED because an error in the loading 
--	library would load Blizzard_DebugTools, confuse LibBuilder, and 
--	seemed to prevent the real error message from showing.
local frame = CreateFrame("Frame")
frame:Hide()
frame:SetScript("OnUpdate", function() LibBuilder:OnUpdate() end)
LibBuilder.frame = frame


--	fields used are listed here for documentation purposes
LibBuilder.workspace = nil
LibBuilder.major = nil
LibBuilder.minor = nil


---	Registers a new library and creates a workspace for it.
--	@param		register	Optional; if true, LibBuilder will attempt to create the library in LibStub. Defaults to true.
--	@param		major		The major version of the library being registered.
--	@param		minor		The minor version of the library being registered.
--	@param		workspace	Optional; an existing table to use as the workspace.
--	@return		workspace	The workspace for the new library.
--	@return		interface	The table returned from LibStub. Only returned if 'register' was not false.
function LibBuilder:NewLib(...)
	local argCount = select("#", ...)
	assert( 2 <= argCount and argCount <= 4, "Invalid number of arguments passed to argCount.")

	local register, major, minor, workspace
	local interface, oldMinor		--	only used if 'register' is not false
	
	--	shift arguments based on the type of the first
	local t = type(...)
	
	if t == "boolean" then
		register, major, minor, workspace = ...
	elseif t == nil then
		register, major, minor, workspace = true, ...
	else
		register, major, minor, workspace = false, ...
	end

	--	check the supplied version info
	if type(major) ~= "string" or type(minor) ~= "number" then
		local message = self:BuildErrorMessage("BADVERSION", major, minor)
		error(message)
	end
	
	if register then
		interface, oldMinor = LibStub:GetLibrary(major, minor)
	end
	
	--	check for collisions
	if self.major or self.minor or self.workspace then
		local message = self:BuildErrorMessage("COLLISION", self.major, self.minor, major, minor)
		
		--	we want the user to see the error but we want don't want to break the new lib
		--	TODO: should I change the code to save the old values, run the normal stuff, 
		--	and then use error?
		geterrorhandler()(message)
	end
	
	--	show the frame so that we can use ADDON_LOADED 
	--	to check if this library was not closed.
	self.frame:Show()

	self.major = major
	self.minor = minor
	
	--	if the programmer gave us a funky workspace, that's their problem
	self.workspace = workspace or {}
	
	if register then
		return self.workspace, interface, oldMinor
	else
		return self.workspace	--	return as a single value rather than three
	end
end


---	Retrieves the version info and workspace for the current active library.
--	@return major		Major version of the library.
--	@return minor		Minor version of the library.
--	@return workspace	The workspace for the library.
function LibBuilder:GetLib()
	if not self.major then		--	just check major to allow for flexibility
		local message = self:BuildErrorMessage("NOLIB", "GetLib")
		error(message)
	end
	
	return self.major, self.minor, self.workspace
end


---	Closes the active library. This is used for tracking, to make sure 
--	that two libraries don't get clobbered together.
--	This function will return the same values as GetLib, as a shortcut 
--	in case the library needs to do last-minute initialization.
--	@see GetLib
function LibBuilder:CloseLib()
	--	we don't want programmers calling this to 
	--	suppress errors before they open their lib.
	if not self.major then
		local message = self:BuildErrorMessage("NOLIB", "CloseLib")
		error(message)
	end
	
	--	the library was closed properly, so disable the OnUpdate frame
	self.frame:Hide()
	
	--	grab values before reset so we can return them
	local major, minor, workspace = self.major, self.minor, self.workspace
	
	self.major, self.minor, self.workspace = nil, nil, nil
	
	return major, minor, workspace
end


---	Private, builds an error message
--	@param	errorName	Name of the error
--	@param	...			List of arguments applicable to the error type.
--	@return message		The complete error string.
function LibBuilder:BuildErrorMessage(errorName, ...)
	local body = self.errorStrings[errorName]:format(...)
	local message = self.errorStrings.SHELL:format(MAJOR, body)

	return message
end


function LibBuilder:OnUpdate()
	--	hide frame so we don't get another update unless another library 
	--	is created, and do it now in case we have to throw the error
	self.frame:Hide()

	if self.major then
		local major, minor = self.major, self.minor 
		
		--	close it so the error only happens once
		self:CloseLib()
		
		local message = self:BuildErrorMessage("LEFTOPEN", major, minor)
		error(message)
	end
end


local function __call(self, ...)
	local argCount = select("#", ...)
	
	if argCount == 0 then
		return self:GetLib()
	elseif argCount == 2 or argCount == 3 then
		return self:NewLib(...)
	end
	
	error("Invalid number of arguments passed to "..MAJOR)
end

setmetatable(LibBuilder, { __call = __call })


if not _G[BASE] then
	local shell = {}
	_G[BASE] = shell
	
	local defaultVersion = "1.0"
	
	local function shellcall(self, ...)
		local argCount = select("#", ...)
		
		--	emtpy arg list -> call GetLib on default version
		--	1 arg -> return that version (or nil)
		--	2 or more -> if first is a field in the shell, get 
		--	that version and pass remaning args, otherwise 
		--	pass all args to default version
		if argCount == 0 then
			return self[defaultVersion]:GetLib()
		elseif argCount == 1 then
			return self[...]
		else
			local arg1 = ...
			if arg1 == "" then	--	empty string is an alias for default
				return self[defaultVersion](select(2, ...))
			if self[arg1] then
				return self[arg1](select(2, ...))
			else
				return self[defaultVersion](...)
			end
		end
	end
	
	setmetatable(shell, { __call = shellcall })
end

_G[MAJOR] = LibBuilder
_G[BASE][VERSION] = LibBuilder
